#!/usr/bin/bash

# vim:ft=sh:fdm=marker:fmr={,}

# archlabs installer library script file
# this file is not meant to be run directly
# sourcing this file in a non bash shell is not advised

declare -gA BCMDS=(
[syslinux]="syslinux-install_update -iam"
[grub]="grub-install --recheck --force"
[systemd-boot]="bootctl --path=/boot install"
)

declare -Agr BMNTS=(
[UEFI-grub]="/boot/efi" [UEFI-systemd-boot]="/boot" [BIOS-grub]="/boot"
[BIOS-syslinux]="/boot" [UEFI-syslinux]="/boot"
)

declare -Agrx BOOTLDRS=(
[BIOS]="grub ${BMNTS[BIOS-grub]} syslinux ${BMNTS[BIOS-syslinux]}"
[UEFI]="systemd-boot ${BMNTS[UEFI-systemd-boot]} grub ${BMNTS[UEFI-grub]} syslinux ${BMNTS[UEFI-syslinux]}"
)

declare -Agx EDIT_FILES=(
[2]="/etc/X11/xorg.conf.d/00-keyboard.conf /etc/default/keyboard /etc/vconsole.conf"
[3]="/etc/locale.conf /etc/default/locale"
[4]="/etc/hostname /etc/hosts"
[5]="/etc/sudoers"
[6]="/etc/mkinitcpio.conf"
[7]="/etc/fstab"
[8]="/etc/crypttab"
[9]="/etc/default/grub"
[10]="/etc/pacman.conf"
[11]=""  # login files, left empty, don't know the username yet for /home/USER/FILES
)

prep_for_grub() {
    local cfg="$MNT/etc/default/grub"
    sed -i "s/GRUB_DISTRIBUTOR=.*/GRUB_DISTRIBUTOR=\"${DIST}\"/g;
            s/GRUB_CMDLINE_LINUX_DEFAULT=.*/GRUB_CMDLINE_LINUX_DEFAULT=\"\"/g" $cfg
    if [[ $LUKS_DEV ]]; then
        sed -i "s~#GRUB_ENABLE_CRYPTODISK~GRUB_ENABLE_CRYPTODISK~g;
                s~GRUB_CMDLINE_LINUX=.*~GRUB_CMDLINE_LINUX=\"${LUKS_DEV}\"~g" $cfg
    fi
    if [[ $SYS == 'BIOS' && $LVM && $SEPERATE_BOOT == false ]]; then
        sed -i "s/GRUB_PRELOAD_MODULES=.*/GRUB_PRELOAD_MODULES=\"lvm\"/g" $cfg
    fi
    return 0
}

prep_for_systemd-boot() {
    # no LVM then systemd-boot uses PARTUUID
    [[ $ROOT_PART =~ /dev/mapper ]] || ROOT_PART_ID="PART$ROOT_PART_ID"

    # create the boot entry configs
    mkdir -p ${MNT}${BMNTS[$SYS-$BOOTLDR]}/loader/entries
    cat > ${MNT}${BMNTS[$SYS-$BOOTLDR]}/loader/loader.conf << EOF
default  $DIST
timeout  5
editor   no
EOF
    cat > ${MNT}${BMNTS[$SYS-$BOOTLDR]}/loader/entries/${DIST}.conf << EOF
title   $DIST Linux
linux   /vmlinuz-${KERNEL}$([[ $UCODE ]] && printf "\ninitrd  %s" "/$UCODE")
initrd  /initramfs-$KERNEL.img
options root=$ROOT_PART_ID $([[ $LUKS_DEV ]] && printf "%s " "$LUKS_DEV")rw
EOF
    # add pacman hook to update the bootloader when systemd receives an update
    mkdir -p $MNT/etc/pacman.d/hooks
    cat > $MNT/etc/pacman.d/hooks/systemd-boot.hook << EOF
[Trigger]
Type = Package
Operation = Upgrade
Target = systemd

[Action]
Description = Updating systemd-boot
When = PostTransaction
Exec = /usr/bin/bootctl update
EOF
    # systemd-boot requires this before running bootctl
    systemd-machine-id-setup --root="$MNT" >/dev/null 2>&1
    return 0
}

prep_for_syslinux() {
    if [[ $SYS == 'UEFI' ]]; then
        local cfgdir="${MNT}${BMNTS[$SYS-$BOOTLDR]}/EFI/syslinux"
        local cfgsrcdir="/usr/lib/syslinux/efi32/"
        [[ $IS_64BIT == true ]] && cfgsrcdir="/usr/lib/syslinux/efi64/"
    else
        local cfgdir="$MNT${BMNTS[$SYS-$BOOTLDR]}/syslinux"
        local cfgsrcdir="/usr/lib/syslinux/bios/"
    fi

    mkdir -p $cfgdir
    cp -r $cfgsrcdir $cfgdir/
    cat > $cfgdir/syslinux.cfg << EOF
UI menu.c32
PROMPT 0

MENU TITLE $DIST Syslinux Boot Menu
TIMEOUT 50
DEFAULT $DIST

LABEL $DIST
MENU LABEL $DIST Linux
LINUX ../vmlinuz-$KERNEL
APPEND root=$ROOT_PART_ID $([[ $LUKS_DEV ]] && printf "%s " "$LUKS_DEV")rw
INITRD ../initramfs-$KERNEL.img$([[ $UCODE ]] && printf "\nINITRD %s" "../$UCODE")

LABEL ${DIST}fallback
MENU LABEL $DIST Linux Fallback
LINUX ../vmlinuz-$KERNEL
APPEND root=$ROOT_PART_ID $([[ $LUKS_DEV ]] && printf "%s " "$LUKS_DEV")rw
INITRD ../initramfs-$KERNEL-fallback.img$([[ $UCODE ]] && printf "\nINITRD %s" "../$UCODE")
EOF
    return 0
}

install_bootloader() {
    # not an LVM we can use the UUID for booting otherwise use the partition label
    if ! [[ $ROOT_PART =~ /dev/mapper ]]; then
        ROOT_PART_ID="UUID=$(blkid -s PARTUUID -o value $ROOT_PART)"
    else
        ROOT_PART_ID="$ROOT_PART"
    fi

    # remove old UEFI boot entries
    if [[ $SYS == 'UEFI' ]]; then
        find ${MNT}${BMNTS[$SYS-$BOOTLDR]}/EFI/ -maxdepth 1 -mindepth 1 \
            -name '[aA][rR][cC][hH][lL]abs' -type d -exec rm -rf '{}' \; >/dev/null 2>&1
    fi

    # sets up the bootloader configs and BCMD (boot command)
    prep_for_$BOOTLDR

    # install/setup the bootloader by running the BCMD (boot command)
    chrun "${BCMDS[$BOOTLDR]}" # 2>$ERR
    echeck "${BCMDS[$BOOTLDR]}"

    # copy efi stub to generic catch all
    [[ $SYS == 'UEFI' && $BOOTLDR =~ (grub|syslinux) ]] && uefi_boot_fallback
    return 0
}

uefi_boot_fallback() {
    # some UEFI firmware requires a directory in the ESP and a generic bootx64.efi
    # see:  https://wiki.archlinux.org/index.php/GRUB#UEFI
    # also: https://wiki.archlinux.org/index.php/syslinux#UEFI_Systems

    local esp="${MNT}${BMNTS[$SYS-$BOOTLDR]}"
    local default="Boot"
    default="$(find $esp/EFI/ -maxdepth 1 -mindepth 1 -name '[Bb][oO][oO][tT]' -type d)"
    [[ $default ]] && default="$(basename $default)"

    # directory exists: remove everything inside, otherwise mkdir it
    [[ -d $esp/EFI/$default ]] && rm -rf $esp/EFI/$default/* || mkdir -p $esp/EFI/$default

    # copy efi stub to the new default location
    if [[ $BOOTLDR == 'syslinux' ]]; then
        # syslinux requires the whole boot setup be copied, not just the stub
        cp -rf $esp/EFI/syslinux/* $esp/EFI/$default/
        cp -f $esp/EFI/syslinux/syslinux.efi $esp/EFI/$default/bootx64.efi
    elif [[ $BOOTLDR == 'grub' ]]; then
        # grub uses different names for the stub depending on architecture
        local ext="ia32"
        [[ $IS_64BIT == true ]] && ext="x64"
        cp -f $esp/EFI/$DIST/grub${ext}.efi $esp/EFI/$default/boot${ext}.efi
    fi
    return 0
}

shim_secure_boot() {
    # still a W.I.P
    local shim_file="shim.efi"
    [[ $IS_64BIT == true ]] && shim_file="shim64.efi"
    efibootmgr -c -w -L $DIST -d $BOOT_DEVICE -p $BOOT_PART_NUM -l ${MNT}${BMNTS[$SYS-$BOOTLDR]}/$shim_file
    return 0
}
